import pytest
from functools import namedtuple

from umongo import Document, Schema, fields


@pytest.fixture
def classroom_model(db):

    class Teacher(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)

        class Config:
            register_document = False
            collection = db.teacher

    class Course(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)
            teacher = fields.ReferenceField(Teacher, required=True)

        class Config:
            register_document = False
            collection = db.course

    class Student(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)
            birthday = fields.DateTimeField()
            courses = fields.ListField(fields.ReferenceField(Course))

        class Config:
            register_document = False
            collection = db.student

    Teacher.collection.drop()
    Course.collection.drop()
    Student.collection.drop()

    return namedtuple('Mapping', ('Teacher', 'Course', 'Student'))(Teacher, Course, Student)



import bson
from marshmallow import validates_schema, ValidationError
from marshmallow import Schema as MaSchema
from marshmallow import fields as ma_fields


# Declare the ObjectIdField here to prevent recursive import error with fields
class ObjectIdField(ma_fields.Field):

    def _serialize(self, value, attr, obj):
        if value is None:
            return None
        return str(value)

    def _deserialize(self, value, attr, data):
        if value is None:
            return None
        try:
            return bson.ObjectId(value)
        except ValueError:
            raise ValidationError('Invalid ObjectId')


__all__ = ('BaseSchema', 'Schema', 'NestedSchema')


class BaseSchema(MaSchema):

    @validates_schema(pass_original=True)
    def check_unknown_fields(self, data, original_data):
        for key in original_data:
            if key not in self.fields:
                raise ValidationError('Unknown field name {}'.format(key))


class Schema(BaseSchema):

    id = ObjectIdField(attribute='_id')


class NestedSchema(BaseSchema):
    pass



from .data_proxy import DataProxy
from .abstract import BaseWrappedData
from .exceptions import NotCreatedError
from .meta import MetaDocument


class Document(BaseWrappedData, metaclass=MetaDocument):

    class Config:
        collection = None
        lazy_collection = None
        register_document = True

    def __init__(self, **kwargs):
        self.created = False
        self.data = DataProxy(self.schema, kwargs)

    @property
    def pk(self):
        return self.data._data.get('_id')

    @property
    def name(self):
        return self.__name__

    @classmethod
    def build_from_mongo(cls, data, partial=False):
        doc = cls()
        doc.from_mongo(data, partial=partial)
        return doc

    def from_mongo(self, data, partial=False):
        # TODO: handle partial
        self.data.from_mongo(data, partial=partial)
        self.created = True

    def to_mongo(self, update=False):
        return self.data.to_mongo(update=update)

    def dump(self):
        return self.data.dump()

    @property
    def collection(self):
        # Cannot implicitly access to the class's property
        return type(self).collection

    @property
    def driver(self):
        # Cannot implicitly access to the class's property
        return type(self).driver

    def reload(self):
        if not self.created:
            raise NotCreatedError('Cannot reload a document that'
                                  ' has not been committed yet')
        return self.driver.reload(self)

    def commit(self, io_validate_all=False):
        return self.driver.commit(self, io_validate_all=False)

    def delete(self):
        return self.driver.delete(self)

    @classmethod
    def find_one(cls, *args, **kwargs):
        return cls.driver.find_one(cls, *args, **kwargs)

    @classmethod
    def find(cls, *args, **kwargs):
        return cls.driver.find(cls, *args, **kwargs)



from bson import ObjectId
import pytest
from datetime import datetime
from marshmallow import ValidationError

from umongo.data_proxy import DataProxy
from umongo import Document, EmbeddedDocument, Schema, NestedSchema, fields


class TestFields:

    def test_datetime(self):

        class MySchema(NestedSchema):
            a = fields.DateTimeField()

        s = MySchema(strict=True)
        data, _ = s.load({'a': datetime(2016, 8, 6)})
        assert data['a'] == datetime(2016, 8, 6)
        data, _ = s.load({'a': "2016-08-06T00:00:00Z"})
        assert data['a'] == datetime(2016, 8, 6)
        with pytest.raises(ValidationError):
            s.load({'a': "2016-08-06"})
        with pytest.raises(ValidationError):
            s.load({'a': "dummy"})

    def test_nested(self):

        class MyNestedSchema(NestedSchema):
            a = fields.IntField(attribute='in_mongo_a')
            b = fields.IntField()

        class MySchema(Schema):
            nested = fields.NestedField(MyNestedSchema(), attribute='in_mongo_nested')

        d = DataProxy(MySchema())
        d.load({'nested': {'a': 1, 'b': 2}})
        with pytest.raises(KeyError):
            d.in_mongo_nested
        assert d.dump() == {'nested': {'a': 1, 'b': 2}}
        assert d.nested == {'in_mongo_a': 1, 'b': 2}
        assert d.to_mongo() == {'in_mongo_nested': {'in_mongo_a': 1, 'b': 2}}

    def test_dict(self):

        class MySchema(Schema):
            dict = fields.DictField(attribute='in_mongo_dict')

        d = DataProxy(MySchema())
        d.load({'dict': {'a': 1, 'b': {'c': True}}})
        with pytest.raises(KeyError):
            d.in_mongo_dict
        assert d.dump() == {'dict': {'a': 1, 'b': {'c': True}}}
        assert d.dict == {'a': 1, 'b': {'c': True}}
        assert d.to_mongo() == {'in_mongo_dict': {'a': 1, 'b': {'c': True}}}
        # TODO: handle check modif of the dict
        # d.dict['a'] = 1
        # assert d.to_mongo(update=True) == {'$set': {'in_mongo_dict': {'a': 1, 'b': {'c': True}}}}

    def test_embedded_document(self):

        class MyEmbeddedDocument(EmbeddedDocument):
            class Schema(Schema):
                a = fields.IntField(attribute='in_mongo_a')
                b = fields.IntField()

        class MySchema(Schema):
            embedded = fields.EmbeddedField(MyEmbeddedDocument, attribute='in_mongo_embedded')

        d = DataProxy(MySchema())
        d.load({'embedded': {'a': 1, 'b': 2}})
        assert d.dump() == {'embedded': {'a': 1, 'b': 2}}
        assert type(d.embedded) == MyEmbeddedDocument
        assert d.embedded.a == 1
        assert d.embedded.b == 2
        assert d.embedded.dump() == {'a': 1, 'b': 2}
        assert d.embedded.to_mongo() == {'in_mongo_a': 1, 'b': 2}
        assert d.to_mongo() == {'in_mongo_embedded': {'in_mongo_a': 1, 'b': 2}}

        d.embedded.a = 3
        assert d.embedded.to_mongo(update=True) == {'$set': {'in_mongo_a': 3}}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_embedded': {'in_mongo_a': 3, 'b': 2}}}

        del d.embedded.a
        assert d.embedded.to_mongo(update=True) == {'$unset': ['in_mongo_a']}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_embedded': {'b': 2}}}

        d.embedded = MyEmbeddedDocument(a=4)
        assert d.embedded.to_mongo(update=True) is None
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_embedded': {'in_mongo_a': 4}}}

    def test_list(self):

        class MySchema(Schema):
            list = fields.ListField(fields.IntField(), attribute='in_mongo_list')

        d = DataProxy(MySchema())
        d.load({'list': [1, 2, 3]})
        assert d.dump() == {'list': [1, 2, 3]}
        assert d.to_mongo() == {'in_mongo_list': [1, 2, 3]}
        assert d.list == [1, 2, 3]
        d.list.append(4)
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_list': [1, 2, 3, 4]}}

        d.clear_modified()
        d.list = [5, 6, 7]
        assert d.dump() == {'list': [5, 6, 7]}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_list': [5, 6, 7]}}

        d.clear_modified()
        d.list.pop()
        assert d.dump() == {'list': [5, 6]}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_list': [5, 6]}}

        d.clear_modified()
        d.list.clear()
        assert d.dump() == {'list': []}
        assert d.to_mongo(update=True) == {'$unset': ['in_mongo_list']}

        d.list = [1, 2, 3]
        d.clear_modified()
        d.list.remove(1)
        assert d.dump() == {'list': [2, 3]}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_list': [2, 3]}}

        d.clear_modified()
        d.list.reverse()
        assert d.dump() == {'list': [3, 2]}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_list': [3, 2]}}

        d.clear_modified()
        d.list.sort()
        assert d.dump() == {'list': [2, 3]}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_list': [2, 3]}}

        d.clear_modified()
        d.list.extend([4, 5])
        assert d.dump() == {'list': [2, 3, 4, 5]}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_list': [2, 3, 4, 5]}}

    def test_objectid(self):

        class MySchema(Schema):
            objid = fields.ObjectIdField(attribute='in_mongo_objid')

        d = DataProxy(MySchema())
        d.load({'objid': ObjectId("5672d47b1d41c88dcd37ef05")})
        assert d.dump() == {'objid': "5672d47b1d41c88dcd37ef05"}
        assert d.to_mongo() == {'in_mongo_objid': ObjectId("5672d47b1d41c88dcd37ef05")}
        d.load({'objid': "5672d47b1d41c88dcd37ef05"})
        assert d.dump() == {'objid': "5672d47b1d41c88dcd37ef05"}
        assert d.to_mongo() == {'in_mongo_objid': ObjectId("5672d47b1d41c88dcd37ef05")}
        assert d.objid == ObjectId("5672d47b1d41c88dcd37ef05")

        d.objid = ObjectId("5672d5e71d41c88f914b77c4")
        d.to_mongo(update=True) == {
            '$set': {'in_mongo_objid': ObjectId("5672d5e71d41c88f914b77c4")}}

        d.objid = ObjectId("5672d5e71d41c88f914b77c4")
        d.to_mongo(update=True) == {
            '$set': {'in_mongo_objid': ObjectId("5672d5e71d41c88f914b77c4")}}

        d.objid = "5672d5e71d41c88f914b77c4"
        assert d.objid == ObjectId("5672d5e71d41c88f914b77c4")

    def test_reference(self):

        class MyReferencedDoc(Document):
            class Schema(Schema):
                pass

        to_refer_doc = MyReferencedDoc.build_from_mongo(
            {'_id': ObjectId("5672d47b1d41c88dcd37ef05")})

        class MySchema(Schema):
            ref = fields.ReferenceField(MyReferencedDoc, attribute='in_mongo_ref')

        d = DataProxy(MySchema())
        d.load({'ref': ObjectId("5672d47b1d41c88dcd37ef05")})
        d.load({'ref': "5672d47b1d41c88dcd37ef05"})
        assert d.dump() == {'ref': "5672d47b1d41c88dcd37ef05"}
        d.ref.document_cls == MyReferencedDoc
        d.ref = to_refer_doc
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_ref': to_refer_doc.pk}}
        assert d.ref.document_cls == MyReferencedDoc

    def test_reference_lazy(self):

        class MyReferencedDocLazy(Document):
            pass

        to_refer_doc = MyReferencedDocLazy.build_from_mongo(
            {'_id': ObjectId("5672d47b1d41c88dcd37ef05")})

        class MySchema(Schema):
            ref = fields.ReferenceField("MyReferencedDocLazy", attribute='in_mongo_ref')

        d = DataProxy(MySchema())
        d.load({'ref': ObjectId("5672d47b1d41c88dcd37ef05")})
        d.load({'ref': "5672d47b1d41c88dcd37ef05"})
        assert d.dump() == {'ref': "5672d47b1d41c88dcd37ef05"}
        assert d.ref.document_cls == MyReferencedDocLazy
        d.ref = to_refer_doc
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_ref': to_refer_doc.pk}}
        assert d.ref.document_cls == MyReferencedDocLazy



import pytest
from bson import ObjectId

from .common import BaseTest
from umongo import (Document, Schema, fields, AlreadyRegisteredDocumentError,
                    NotRegisteredDocumentError)


class TestRegisterDocuments(BaseTest):

    def test_already_register_documents(self):

        class Doc(Document):
            pass

        with pytest.raises(AlreadyRegisteredDocumentError):

            class Doc(Document):
                pass

    def test_not_register_documents(self):

        class Doc(Document):

            class Schema(Schema):
                ref = fields.ReferenceField('DummyDoc')

        with pytest.raises(NotRegisteredDocumentError):
            Doc(ref=ObjectId('56dee8dd1d41c8860b263d86'))

    def test_dont_register_documents(self):

        class Doc(Document):
            pass

        # Should not raise `AlreadyRegisteredDocumentError`
        class Doc(Document):

            class Config:
                register_document = False



from .abstract import AbstractCursor


class Cursor(AbstractCursor):

    def __init__(self, document_cls, cursor, *args, **kwargs):
        self.__dict__['raw_cursor'] = cursor
        self.__dict__['document_cls'] = document_cls
        super().__init__(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.__dict__['raw_cursor'], name)

    def __setattr__(self, name, value):
        return setattr(self.raw_cursor, name, value)

    def __next__(self):
        elem = next(self.raw_cursor)
        return self.document_cls.build_from_mongo(elem)

    def __iter__(self):
        for elem in self.raw_cursor:
            yield self.document_cls.build_from_mongo(elem)



from umongo.registerer import default_registerer


class BaseTest:

    def setup(self):
        default_registerer.documents = {}


def get_pymongo_version():
    import pymongo
    version = getattr(pymongo, '__version__', None) or getattr(pymongo, 'version')
    version = [int(i) for i in version.split('.')]
    if len(version) == 2:
        version.append(0)
    return version



import pytest
import asyncio
from datetime import datetime
from functools import namedtuple

# Check if the required dependancies are met to run this driver's tests
try:
    from motor.motor_asyncio import AsyncIOMotorClient
except ImportError as e:
    dep_error = str(e)
else:
    dep_error = None

from ..common import BaseTest

from umongo import Document, Schema, fields

if not dep_error:  # Make sure the module is valid by importing it
    from umongo.dal import motor_asyncio


@pytest.fixture
def db():
    return AsyncIOMotorClient().umongo_test


@pytest.fixture
def classroom_model(db):

    class Teacher(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)

        class Config:
            register_document = False
            collection = db.teacher

    class Course(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)
            teacher = fields.ReferenceField(Teacher, required=True)

        class Config:
            register_document = False
            collection = db.course

    class Student(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)
            birthday = fields.DateTimeField()
            courses = fields.ListField(fields.ReferenceField(Course))

        class Config:
            register_document = False
            collection = db.student

    loop = asyncio.get_event_loop()
    loop.set_debug(True)

    @asyncio.coroutine
    def clean_db():
        yield from Teacher.collection.drop()
        yield from Course.collection.drop()
        yield from Student.collection.drop()

    loop.run_until_complete(clean_db())

    return namedtuple('Mapping', ('Teacher', 'Course', 'Student'))(Teacher, Course, Student)


@pytest.fixture
def loop():
    return asyncio.get_event_loop()


@pytest.mark.skipif(dep_error is not None, reason=dep_error)
class TestMotorAsyncio(BaseTest):

    def test_create(self, loop, db):

        @asyncio.coroutine
        def do_test():
            class Student(Document):

                class Schema(Schema):
                    name = fields.StrField(required=True)
                    birthday = fields.DateTimeField()

                class Config:
                    collection = db.student

            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
            yield from john.commit()
            assert john.to_mongo() == {
                '_id': john.data.id,
                'name': 'John Doe',
                'birthday': datetime(1995, 12, 12)
            }

            john2 = yield from Student.find_one(john.data.id)
            assert john2.data == john.data

        loop.run_until_complete(do_test())

    def test_update(self, loop, db):

        @asyncio.coroutine
        def do_test():

            class Student(Document):

                class Schema(Schema):
                    name = fields.StrField(required=True)
                    birthday = fields.DateTimeField()

                class Config:
                    collection = db.student

            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
            yield from john.commit()
            john.data.name = 'William Doe'
            assert john.to_mongo(update=True) == {'$set': {'name': 'William Doe'}}
            yield from john.commit()
            assert john.to_mongo(update=True) == None
            john2 = yield from Student.find_one(john.data.id)
            assert john2.data == john.data

        loop.run_until_complete(do_test())

    def test_reload(self, loop, db):

        @asyncio.coroutine
        def do_test():

            class Student(Document):

                class Schema(Schema):
                    name = fields.StrField(required=True)
                    birthday = fields.DateTimeField()

                class Config:
                    collection = db.student

            john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
            yield from john.commit()
            john2 = yield from Student.find_one(john.data.id)
            john2.data.name = 'William Doe'
            yield from john2.commit()
            yield from john.reload()
            assert john.data.name == 'William Doe'

        loop.run_until_complete(do_test())

    @pytest.mark.xfail
    def test_cusor(self, loop, db):

        @asyncio.coroutine
        def do_test():

            class Student(Document):

                class Schema(Schema):
                    name = fields.StrField(required=True)
                    birthday = fields.DateTimeField()

                class Config:
                    collection = db.student

            Student.collection.drop()

            for i in range(10):
                yield from Student(name='student-%s' % i).commit()
            cursor = yield from Student.find(limit=5, skip=6)
            assert cursor.count() == 10
            assert cursor.count(with_limit_and_skip=True) == 4
            names = []
            for elem in cursor:
                assert isinstance(elem, Student)
                names.append(elem.data.name)
            assert sorted(names) == ['student-%s' % i for i in range(6, 10)]

        loop.run_until_complete(do_test())

    def test_classroom(self, loop, classroom_model):

        @asyncio.coroutine
        def do_test():

            student = classroom_model.Student(name='Marty McFly', birthday=datetime(1968, 6, 9))
            yield from student.commit()
            teacher = classroom_model.Teacher(name='M. Strickland')
            yield from teacher.commit()
            course = classroom_model.Course(name='Overboard 101', teacher=teacher)
            yield from course.commit()
            assert student.data.courses == []
            student.data.courses.append(course)
            yield from student.commit()
            assert student.to_mongo() == {
                '_id': student.pk,
                'name': 'Marty McFly',
                'birthday': datetime(1968, 6, 9),
                'courses': [course.pk]
            }

        loop.run_until_complete(do_test())



from marshmallow import ValidationError
import pytest

from umongo.data_proxy import DataProxy
from umongo import Schema, fields


class TestDataProxy:

    def test_simple(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField()

        d = DataProxy(MySchema())
        d.load({'a': 1, 'b': 2})
        assert d.a == 1
        d.b = 3
        assert d.b == 3
        assert d._data == {'a': 1, 'b': 3}
        d['a'] = 4
        assert d['a'] == 4
        assert d.a == d['a']
        assert d.dump() == {'a': 4, 'b': 3}
        del d.b
        assert d.dump() == {'a': 4}

    def test_load(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        d = DataProxy(MySchema())
        d.load({'a': 1, 'b': 2})
        assert d.to_mongo() == {'a': 1, 'in_mongo_b': 2}

        d.a = 3
        assert d.to_mongo(update=True) == {'$set': {'a': 3}}

        d.load({'a': 4, 'b': 5})
        assert d.to_mongo(update=True) is None
        assert d.to_mongo() == {'a': 4, 'in_mongo_b': 5}

    def test_modify(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        d = DataProxy(MySchema())
        d.load({'a': 1, 'b': 2})
        assert d.to_mongo() == {'a': 1, 'in_mongo_b': 2}
        assert d.to_mongo(update=True) is None
        d.a = 3
        del d.b
        assert d.to_mongo(update=True) == {'$set': {'a': 3}, '$unset': ['in_mongo_b']}
        d.clear_modified()
        assert d.to_mongo(update=True) is None
        assert d.to_mongo() == {'a': 3}

    def test_set(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        d = DataProxy(MySchema())
        d.load({'a': 1, 'b': 2})
        d.a = 3
        assert d.to_mongo() == {'a': 3, 'in_mongo_b': 2}
        assert d.to_mongo(update=True) == {'$set': {'a': 3}}

        d.load({'a': 1, 'b': 2})
        d.b = 3
        assert d.to_mongo() == {'a': 1, 'in_mongo_b': 3}
        assert d.to_mongo(update=True) == {'$set': {'in_mongo_b': 3}}

    def test_del(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField(attribute='in_mongo_b')

        d = DataProxy(MySchema())
        d.load({'a': 1, 'b': 2})
        del d.b
        assert d.to_mongo() == {'a': 1}
        assert d.to_mongo(update=True) == {'$unset': ['in_mongo_b']}
        del d.a
        assert d.to_mongo(update=True) == {'$unset': ['a', 'in_mongo_b']}

    def test_item(self):

        class MySchema(Schema):
            a = fields.IntField()
            b = fields.IntField()

        d = DataProxy(MySchema())
        d.load({'a': 1, 'b': 2})
        d['a'] = 2
        assert d.to_mongo() == {'a': 2, 'b': 2}
        assert d.to_mongo(update=True) == {'$set': {'a': 2}}
        assert d['a'] == d.a == 2
        del d['a']
        assert d.to_mongo() == {'b': 2}
        assert d.to_mongo(update=True) == {'$unset': ['a']}

    def test_route_naming(self):

        class MySchema(Schema):
            in_front = fields.IntField(attribute='in_mongo')

        d = DataProxy(MySchema())
        with pytest.raises(ValidationError):
            d.load({'in_mongo': 42})
        d.load({'in_front': 42})
        with pytest.raises(KeyError):
            d.in_mongo
        assert d.in_front == 42
        d.in_front = 24
        assert d._data == {'in_mongo': 24}
        assert d.in_front == 24
        assert d.dump() == {'in_front': 24}
        assert d.to_mongo() == {'in_mongo': 24}

    def test_from_mongo(self):

        class MySchema(Schema):
            in_front = fields.IntField(attribute='in_mongo')

        d = DataProxy(MySchema())
        with pytest.raises(KeyError):
            d.from_mongo({'in_front': 42})
        d.from_mongo({'in_mongo': 42})
        assert d.in_front == 42



from marshmallow import ValidationError, missing
from functools import partial

from .abstract import BaseWrappedData, ChangeTracker
from .exceptions import FieldNotLoadedError


class DataProxy(BaseWrappedData):

    def __init__(self, schema, data=None):
        self.__dict__['partial'] = False
        self.__dict__['_schema'] = schema
        self.__dict__['_fields'] = schema.fields
        self.__dict__['_data'] = {}
        self.__dict__['_modified_data'] = set()
        fields_from_mongo_key = {}
        for k, v in self._fields.items():
            if v.attribute:
                k = v.attribute
            fields_from_mongo_key[k] = v
        self.__dict__['_fields_from_mongo_key'] = fields_from_mongo_key
        if data:
            self.load(data)

    def to_mongo(self, update=False):
        if update:
            return self._to_mongo_update()
        mongo_data = {}
        for k, v in self._data.items():
            if v is not missing:
                field = self._fields_from_mongo_key[k]
                if hasattr(field, '_serialize_to_mongo'):
                    v = field._serialize_to_mongo(v)
                if v is not missing:
                    mongo_data[k] = v
        return mongo_data

    def _to_mongo_update(self):
        mongo_data = {}
        if self._modified_data:
            set_data = {}
            unset_data = []
            for k in self._modified_data:
                v = self._data[k]
                field = self._fields_from_mongo_key[k]
                if v is not missing and hasattr(field, '_serialize_to_mongo'):
                    v = field._serialize_to_mongo(v)
                if v is missing:
                    unset_data.append(k)
                else:
                    set_data[k] = v
            if set_data:
                mongo_data['$set'] = set_data
            if unset_data:
                mongo_data['$unset'] = sorted(unset_data)
        return mongo_data or None

    def from_mongo(self, data, partial=False):
        self._data = {}
        for k, v in data.items():
            field = self._fields_from_mongo_key[k]
            if hasattr(field, '_deserialize_from_mongo'):
                v = field._deserialize_from_mongo(v)
            self._data[k] = v
        self.clear_modified()
        self._connect_data_change_trackers()
        self.partial = partial

    def dump(self):
        data, err = self._schema.dump(self._data)
        if err:
            raise ValidationError(err)
        return data

    def _mark_as_modified(self, key):
        self._modified_data.add(key)

    def _connect_data_change_trackers(self):
        for k, v in self._data.items():
            if isinstance(v, ChangeTracker):
                v.change_tracker_connect(partial(self._mark_as_modified, k))

    def load(self, data, partial=False):
        data, err = self._schema.load(data)
        if err:
            raise ValidationError(err)
        self._data = data
        self._connect_data_change_trackers()
        self.clear_modified()
        self.partial = partial

    def get_by_mongo_name(self, name):
        return self._data[name]

    def set_by_mongo_name(self, name, value):
        self._data[name] = value
        self._mark_as_modified(name)

    def del_by_mongo_name(self, name):
        self.set_by_mongo_name(name, missing)

    def __getitem__(self, name):
        if name not in self._fields:
            raise KeyError(name)
        name = self._fields[name].attribute or name
        value = self._data.get(name)
        if value is missing:
            if self.partial:
                raise FieldNotLoadedError(name)
            else:
                return None
        return value

    def __delitem__(self, name):
        if name not in self._fields:
            raise KeyError(name)
        name = self._fields[name].attribute or name
        if self._data[name] is missing:
            raise KeyError(name)
        self._data[name] = missing
        self._mark_as_modified(name)

    def __setitem__(self, name, value):
        if name not in self._fields:
            raise KeyError(name)
        field = self._fields[name]
        name = field.attribute or name
        value = field._deserialize(value, name, None)
        self._data[name] = value
        if isinstance(value, ChangeTracker):
            value.change_tracker_connect(partial(self._mark_as_modified, name))
        self._mark_as_modified(name)

    def __setattr__(self, name, value):
        if name in self.__dict__:
            self.__dict__[name] = value
        else:
            self[name] = value

    def __getattr__(self, name):
        return self[name]

    def __delattr__(self, name):
        del self[name]

    def __repr__(self):
        return "<DataProxy(%s)>" % self._data

    def __eq__(self, other):
        if isinstance(other, dict):
            return self._data == other
        else:
            return self._data == other._data

    def clear_modified(self):
        self._modified_data.clear()

    def io_validate(self, validate_all=False):
        # TODO: handle required here
        # TODO: handle unique here
        pass



import pytest
from datetime import datetime
from functools import namedtuple, wraps

from ..common import BaseTest, get_pymongo_version

# Check if the required dependancies are met to run this driver's tests
try:
    from txmongo import MongoConnection
    major, minor, _ = get_pymongo_version()
    if major != 3 or minor < 2:
        dep_error = "txmongo requires pymongo>=3.2.0"
except ImportError:
    dep_error = 'Missing txmongo module'
    # Given the test function are generator, we must wrap them into a dummy
    # function that pytest can skip
    def skip_wrapper(f):

        @wraps(f)
        def wrapper(self):
            pass

        return wrapper

    pytest_inlineCallbacks = skip_wrapper
else:
    dep_error = None
    pytest_inlineCallbacks = pytest.inlineCallbacks

from umongo import Document, Schema, fields


if not dep_error:  # Make sure the module is valid by importing it
    from umongo.dal import txmongo


@pytest.fixture
def db():
    return MongoConnection().umongo_test


@pytest.fixture
def classroom_model(db):

    class Teacher(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)

        class Config:
            register_document = False
            collection = db.teacher

    class Course(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)
            teacher = fields.ReferenceField(Teacher, required=True)

        class Config:
            register_document = False
            collection = db.course

    class Student(Document):

        class Schema(Schema):
            name = fields.StrField(required=True)
            birthday = fields.DateTimeField()
            courses = fields.ListField(fields.ReferenceField(Course))

        class Config:
            register_document = False
            collection = db.student

    pytest.blockon(Teacher.collection.drop())
    pytest.blockon(Course.collection.drop())
    pytest.blockon(Student.collection.drop())

    return namedtuple('Mapping', ('Teacher', 'Course', 'Student'))(Teacher, Course, Student)


@pytest.mark.skipif(dep_error is not None, reason=dep_error)
class TestTxMongo(BaseTest):

    @pytest_inlineCallbacks
    def test_create(self, db):

        class Student(Document):

            class Schema(Schema):
                name = fields.StrField(required=True)
                birthday = fields.DateTimeField()

            class Config:
                collection = db.student

        john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
        yield john.commit()
        assert john.to_mongo() == {
            '_id': john.data.id,
            'name': 'John Doe',
            'birthday': datetime(1995, 12, 12)
        }

        john2 = yield Student.find_one(john.data.id)
        assert john2.data == john.data

    @pytest_inlineCallbacks
    def test_update(self, db):

        class Student(Document):

            class Schema(Schema):
                name = fields.StrField(required=True)
                birthday = fields.DateTimeField()

            class Config:
                collection = db.student

        john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
        yield john.commit()
        john.data.name = 'William Doe'
        assert john.to_mongo(update=True) == {'$set': {'name': 'William Doe'}}
        yield john.commit()
        assert john.to_mongo(update=True) == None
        john2 = yield Student.find_one(john.data.id)
        assert john2.data == john.data

    @pytest_inlineCallbacks
    def test_reload(self, db):

        class Student(Document):

            class Schema(Schema):
                name = fields.StrField(required=True)
                birthday = fields.DateTimeField()

            class Config:
                collection = db.student

        john = Student(name='John Doe', birthday=datetime(1995, 12, 12))
        yield john.commit()
        john2 = yield Student.find_one(john.data.id)
        john2.data.name = 'William Doe'
        yield john2.commit()
        yield john.reload()
        assert john.data.name == 'William Doe'

    @pytest_inlineCallbacks
    def test_cusor(self, db):

        class Student(Document):

            class Schema(Schema):
                name = fields.StrField(required=True)
                birthday = fields.DateTimeField()

            class Config:
                collection = db.student

        Student.collection.drop()

        for i in range(10):
            yield Student(name='student-%s' % i).commit()
        cursor = yield Student.find(limit=5, skip=6)
        assert cursor.count() == 10
        assert cursor.count(with_limit_and_skip=True) == 4
        names = []
        for elem in cursor:
            assert isinstance(elem, Student)
            names.append(elem.data.name)
        assert sorted(names) == ['student-%s' % i for i in range(6, 10)]

    @pytest_inlineCallbacks
    def test_classroom(self, classroom_model):
        student = classroom_model.Student(name='Marty McFly', birthday=datetime(1968, 6, 9))
        yield student.commit()
        teacher = classroom_model.Teacher(name='M. Strickland')
        yield teacher.commit()
        course = classroom_model.Course(name='Overboard 101', teacher=teacher)
        yield course.commit()
        assert student.data.courses == []
        student.data.courses.append(course)
        yield student.commit()
        assert student.to_mongo() == {
            '_id': student.pk,
            'name': 'Marty McFly',
            'birthday': datetime(1968, 6, 9),
            'courses': [course.pk]
        }



from marshmallow.fields import Field

from .registerer import register_document
from .exceptions import NoCollectionDefinedError
from .schema import Schema
from .dal import find_dal_from_collection


def base_meta_document(name, bases, nmspc):
    # Retrieve Schema
    schema_cls = nmspc.get('Schema')
    if not schema_cls:
        schema_cls = next((getattr(base, 'Schema') for base in bases
                           if hasattr(base, 'Schema')), Schema)
    # Retrieve fields declared inside the class
    schema_nmspc = {}
    for key, item in nmspc.items():
        if isinstance(item, Field):
            schema_nmspc[key] = item
            nmspc.pop(key)
    # Need to create a custom Schema class to use the provided fields
    if schema_nmspc:
        schema_cls = type('SubSchema', (schema_cls, ), schema_nmspc)
        nmspc['Schema'] = schema_cls
    # Create Schema instance if not provided
    if 'schema' not in nmspc:
        nmspc['schema'] = schema_cls()
    return name, bases, nmspc


class MetaEmbeddedDocument(type):

    def __new__(cls, name, bases, nmspc):
        name, bases, nmspc = base_meta_document(name, bases, nmspc)
        gen_cls = type.__new__(cls, name, bases, nmspc)
        return gen_cls


class MetaDocument(type):

    def __new__(cls, name, bases, nmspc):
        name, bases, nmspc = base_meta_document(name, bases, nmspc)
        nmspc['_collection'] = None
        nmspc['_driver'] = None
        # Create config with inheritance
        if 'config' not in nmspc:
            config = {}
            for base in bases:
                config.update(getattr(base, 'config', {}))
            config_cls = nmspc.get('Config')
            if config_cls:
                config.update({k: v for k, v in config_cls.__dict__.items()
                               if not k.startswith('_')})
            nmspc['config'] = config
        # Finally create the Document class and register it
        gen_cls = type.__new__(cls, name, bases, nmspc)
        if gen_cls.config.get('register_document'):
            register_document(gen_cls)
        return gen_cls

    @property
    def driver(self):
        if not self._driver:
            self._driver = find_dal_from_collection(self.collection)
        return self._driver

    @property
    def collection(self):
        # This is defined inside the metaclass to give property
        # access to the generated class
        if not self._collection:
            self._collection = self.config.get('collection')
            if not self._collection:
                lazy_collection = self.config.get('lazy_collection')
                if not lazy_collection:
                    raise NoCollectionDefinedError("Not collection nor lazy_collection defined")
                self._collection = lazy_collection()
                if not self._collection:
                    raise NoCollectionDefinedError("lazy_collection didn't returned a collection")
        return self._collection



from datetime import datetime
from marshmallow import ValidationError, missing
from marshmallow import fields as ma_fields

from .registerer import retrieve_document
from .wrapped_data import Reference, List


__all__ = (
    'FieldField',
    'RawField',
    'NestedField',
    'DictField',
    'ListField',
    'StringField',
    'UUIDField',
    'NumberField',
    'IntegerField',
    'DecimalField',
    'BooleanField',
    'FormattedStringField',
    'FloatField',
    'DateTimeField',
    'TimeField',
    'DateField',
    'TimeDeltaField',
    'UrlField',
    'URLField',
    'EmailField',
    'MethodField',
    'FunctionField',
    'StrField',
    'BoolField',
    'IntField',
    'ConstantField',
    'ObjectIdField',
    'ReferenceField',
    'EmbeddedField'
)


# Republish supported mashmallow fields


class FieldField(ma_fields.Field):
    pass


class RawField(ma_fields.Raw):
    pass


class NestedField(ma_fields.Nested):
    pass


class DictField(ma_fields.Dict):
    pass


class ListField(ma_fields.List):

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('default', [])
        kwargs.setdefault('missing', list)
        super().__init__(*args, **kwargs)

    def _deserialize(self, value, attr, data):
        value = super()._deserialize(value, attr, data)
        return self._deserialize_from_mongo(value)

    def _serialize_to_mongo(self, value):
        if not value:
            return missing
        if hasattr(self.container, '_serialize_to_mongo'):
            return [self.container._serialize_to_mongo(each) for each in value]
        else:
            return list(value)

    def _deserialize_from_mongo(self, value):
        return List(value) if value else List()


class StringField(ma_fields.String):
    pass


class UUIDField(ma_fields.UUID):
    pass


class NumberField(ma_fields.Number):
    pass


class IntegerField(ma_fields.Integer):
    pass


class DecimalField(ma_fields.Decimal):
    pass


class BooleanField(ma_fields.Boolean):
    pass


class FormattedStringField(ma_fields.FormattedString):
    pass


class FloatField(ma_fields.Float):
    pass


class DateTimeField(ma_fields.DateTime):
    def _deserialize(self, value, attr, data):
        if isinstance(value, datetime):
            return value
        return super()._deserialize(value, attr, data)


class LocalDateTimeField(ma_fields.LocalDateTime):
    pass


class TimeField(ma_fields.Time):
    pass


class DateField(ma_fields.Date):
    pass


class TimeDeltaField(ma_fields.TimeDelta):
    pass


class UrlField(ma_fields.Url):
    pass


class URLField(ma_fields.URL):
    pass


class EmailField(ma_fields.Email):
    pass


class MethodField(ma_fields.Method):
    pass


class FunctionField(ma_fields.Function):
    pass


class StrField(ma_fields.Str):
    pass


class BoolField(ma_fields.Bool):
    pass


class IntField(ma_fields.Int):
    pass


class ConstantField(ma_fields.Constant):
    pass


# Bonus: new fields !


# ObjectIdField is declared in schema to prevent recursive import error
from .schema import ObjectIdField  # noqa


class ReferenceField(ObjectIdField):

    def __init__(self, document_cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._document_cls = document_cls

    @property
    def document_cls(self):
        if isinstance(self._document_cls, str):
            self._document_cls = retrieve_document(self._document_cls)
        return self._document_cls

    def _serialize(self, value, attr, obj):
        return super()._serialize(value.pk, attr, obj)

    def _deserialize(self, value, attr, data):
        if value is None:
            return None
        from .document import Document
        if isinstance(value, Reference):
            if value.document_cls != self.document_cls:
                raise ValidationError("%s reference expected" % self.document_cls.__name__)
            return value
        elif isinstance(value, self.document_cls):
            if not value.created:
                raise ValidationError("Cannot reference a document that has not been created yet")
            value = value.pk
        elif isinstance(value, Document):
            raise ValidationError("%s reference expected" % self.document_cls.__name__)
        value = super()._deserialize(value, attr, data)
        return self._deserialize_from_mongo(value)

    def _serialize_to_mongo(self, value):
        return value.pk

    def _deserialize_from_mongo(self, value):
        return Reference(self.document_cls, value)

    def io_validate(self):
        pass  # TODO


class EmbeddedField(ma_fields.Dict):

    def __init__(self, embedded_document_cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._embedded_document_cls = embedded_document_cls

    def _serialize(self, value, attr, obj):
        return value.dump()

    def _deserialize(self, value, attr, data):
        if isinstance(value, self._embedded_document_cls):
            return value
        value = super()._deserialize(value, attr, data)
        return self._deserialize_from_mongo(value)

    def _serialize_to_mongo(self, value):
        return value.to_mongo()

    def _deserialize_from_mongo(self, value):
        embedded_document = self._embedded_document_cls()
        embedded_document.load(value)
        return embedded_document



from motor.motor_asyncio import AsyncIOMotorCollection

from ..abstract import AbstractDal
from ..data_proxy import DataProxy
from ..exceptions import NotCreatedError, UpdateError


class MotorAsyncIODal(AbstractDal):

    @staticmethod
    def is_compatible_with(collection):
        return isinstance(collection, AsyncIOMotorCollection)

    def reload(self, doc):
        ret = yield from doc.collection.find_one(doc.pk)
        if ret is None:
            raise NotCreatedError("Document doesn't exists in database")
        doc.data = DataProxy(doc.schema)
        doc.data.from_mongo(ret)

    def commit(self, doc, io_validate_all=False):
        doc.data.io_validate(validate_all=io_validate_all)
        payload = doc.data.to_mongo(update=doc.created)
        if doc.created:
            if payload:
                ret = yield from doc.collection.update(
                    {'_id': doc.data.get_by_mongo_name('_id')}, payload)
                if ret.get('nModified') != 1:
                    raise UpdateError(ret.raw_result)
        else:
            ret = yield from doc.collection.insert(payload)
            # TODO: check ret ?
            doc.data.set_by_mongo_name('_id', ret)
            doc.created = True
        doc.data.clear_modified()

    def delete(self, doc):
        raise NotImplementedError()

    def find_one(self, doc_cls, *args, **kwargs):
        ret = yield from doc_cls.collection.find_one(*args, **kwargs)
        if ret is not None:
            ret = doc_cls.build_from_mongo(ret)
        return ret

    def find(self, doc_cls, *args, **kwargs):
        from ..cursor import Cursor
        raw_cursor = yield from doc_cls.collection.find(*args, **kwargs)
        return Cursor(doc_cls, raw_cursor)



from .exceptions import AlreadyRegisteredDocumentError, NotRegisteredDocumentError


class DocumentRegisterer:
    def __init__(self):
        self.documents = {}

    def register(self, doc):
        name = doc.__name__
        previous_doc = self.documents.get(name)
        if previous_doc:
            raise AlreadyRegisteredDocumentError(
                'Document `%s` has already registerer as %r' % (name, previous_doc))
        self.documents[name] = doc

    def retrieve(self, name):
        doc = self.documents.get(name)
        if not doc:
            raise NotRegisteredDocumentError(
                "Document `%s` hasn't been registered" % name)
        return doc


default_registerer = DocumentRegisterer()
register_document = default_registerer.register
retrieve_document = default_registerer.retrieve



from txmongo.collection import Collection
from twisted.internet.defer import inlineCallbacks

from ..abstract import AbstractCursor, AbstractDal
from ..data_proxy import DataProxy
from ..exceptions import NotCreatedError, UpdateError


class TxCursor(AbstractCursor):

    def __init__(self, document_cls, cursor, *args, **kwargs):
        self._next_index = 0
        self.raw_cursor = cursor
        self.document_cls = document_cls
        self._total = len(cursor)
        self._total_with_limit_and_skip = len(cursor)

    def __next__(self):
        if self._next_index >= self._total_with_limit_and_skip:
            raise StopIteration
        else:
            val = self.raw_cursor[self._next_index]
            self._next_index += 1
            return val

    def __iter__(self):
        for elem in self.raw_cursor:
            yield self.document_cls.build_from_mongo(elem)

    def count(self, with_limit_and_skip=False):
        if with_limit_and_skip:
            return self._total
        else:
            return 10  # !!!


class TxMongoDal(AbstractDal):

    @staticmethod
    def is_compatible_with(collection):
        return isinstance(collection, Collection)

    @inlineCallbacks
    def reload(self, doc):
        ret = yield doc.collection.find_one(doc.pk)
        if ret is None:
            raise NotCreatedError("Document doesn't exists in database")
        doc.data = DataProxy(doc.schema)
        doc.data.from_mongo(ret)

    @inlineCallbacks
    def commit(self, doc, io_validate_all=False):
        doc.data.io_validate(validate_all=io_validate_all)
        payload = doc.data.to_mongo(update=doc.created)
        if doc.created:
            if payload:
                ret = yield doc.collection.update_one(
                    {'_id': doc.data.get_by_mongo_name('_id')}, payload)
                if ret.modified_count != 1:
                    raise UpdateError(ret.raw_result)
        else:
            ret = yield doc.collection.insert_one(payload)
            # TODO: check ret ?
            doc.data.set_by_mongo_name('_id', ret.inserted_id)
            doc.created = True
        doc.data.clear_modified()

    def delete(self, doc):
        raise NotImplementedError()

    @inlineCallbacks
    def find_one(self, doc_cls, *args, **kwargs):
        ret = yield doc_cls.collection.find_one(*args, **kwargs)
        if ret is not None:
            ret = doc_cls.build_from_mongo(ret)
        return ret

    @inlineCallbacks
    def find(self, doc_cls, *args, **kwargs):
        raw_cursor = yield doc_cls.collection.find(*args, **kwargs)
        return TxCursor(doc_cls, raw_cursor)



from .document import Document
from .exceptions import (
    UMongoError,
    ValidationError,
    NoDBDefinedError,
    NotRegisteredDocumentError,
    AlreadyRegisteredDocumentError,
    SchemaFieldNamingClashError,
    UpdateError,
    MissingSchemaError,
    NotCreatedError,
    NoCollectionDefinedError,
    FieldNotLoadedError
)
from . import fields
from .schema import BaseSchema, Schema, NestedSchema
from .wrapped_data import EmbeddedDocument


__author__ = 'Emmanuel Leblond'
__email__ = 'emmanuel.leblond@gmail.com'
__version__ = '0.4.1'
__all__ = (
    'Document',
    'EmbeddedDocument',

    'UMongoError',
    'ValidationError',
    'NoDBDefinedError',
    'NotRegisteredDocumentError',
    'AlreadyRegisteredDocumentError',
    'SchemaFieldNamingClashError',
    'UpdateError',
    'MissingSchemaError',
    'NotCreatedError',
    'NoCollectionDefinedError',
    'FieldNotLoadedError',

    'fields',

    'BaseSchema',
    'Schema',
    'NestedSchema'
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup, find_packages
except ImportError:
    from distutils.core import setup, find_packages


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    "marshmallow>=2.6.0",
    "pymongo"  # needed for bson module
]

test_requirements = [
    "txmongo",
    "motor>=0.6.0"
]

setup(
    name='umongo',
    version='0.4.1',
    description="Small but efficient MongoDB ODM",
    long_description=readme + '\n\n' + history,
    author="Emmanuel Leblond",
    author_email='emmanuel.leblond@gmail.com',
    url='https://github.com/touilleMan/umongo',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    license="MIT",
    zip_safe=False,
    keywords='umongo',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



from marshmallow import ValidationError


class UMongoError(Exception):
    pass


class ValidationError(ValidationError, UMongoError):
    pass


class NoDBDefinedError(UMongoError):
    pass


class NotRegisteredDocumentError(UMongoError):
    pass


class AlreadyRegisteredDocumentError(UMongoError):
    pass


class SchemaFieldNamingClashError(UMongoError):
    pass


class UpdateError(UMongoError):
    pass


class DeleteError(UMongoError):
    pass


class MissingSchemaError(UMongoError):
    pass


class NotCreatedError(UMongoError):
    pass


class NoCollectionDefinedError(UMongoError):
    pass


class FieldNotLoadedError(UMongoError):
    pass


class NoCompatibleDal(UMongoError):
    pass



"""
dal
===

Driver Abstraction Layer
"""

from ..exceptions import NoCompatibleDal


class DalRegisterer:

    def __init__(self):
        self._dals = []

    def register(self, dal):
        self._dals.append(dal)

    def find_from_collection(self, collection):
        for dal in self._dals:
            if dal.is_compatible_with(collection):
                return dal
        raise NoCompatibleDal('Cannot find a umongo dal compatible with %s' %
                              type(collection))


default_dal_registerer = DalRegisterer()
register_dal = default_dal_registerer.register
find_dal_from_collection = default_dal_registerer.find_from_collection

# try to load all the dals by default

try:
    from .pymongo import PyMongoDal
    register_dal(PyMongoDal())
except ImportError:
    pass
try:
    from .txmongo import TxMongoDal
    register_dal(TxMongoDal())
except ImportError:
    pass
try:
    from .motor_asyncio import MotorAsyncIODal
    register_dal(MotorAsyncIODal())
except ImportError:
    pass
try:
    from .motor_tornado import MotorTornadoDal
    register_dal(MotorTornadoDal())
except ImportError:
    pass



from datetime import datetime

from .common import BaseTest
from umongo import Document, Schema, fields


class Student(Document):
    class Schema(Schema):
        name = fields.StrField(required=True)
        birthday = fields.DateTimeField()
        gpa = fields.FloatField()


class TestDocument(BaseTest):

    def test_create(self):
        john = Student(name='John Doe', birthday=datetime(1995, 12, 12), gpa=3.0)
        assert john.to_mongo() == {
            'name': 'John Doe',
            'birthday': datetime(1995, 12, 12),
            'gpa': 3.0
        }
        assert john.created is False

    def test_from_mongo(self):
        john = Student.build_from_mongo(data={
            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
        assert john.to_mongo(update=True) is None
        assert john.created is True
        assert john.to_mongo() == {
            'name': 'John Doe',
            'birthday': datetime(1995, 12, 12),
            'gpa': 3.0
        }

    def test_update(self):
        john = Student.build_from_mongo(data={
            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
        john.data.name = 'William Doe'
        john.data.birthday = datetime(1996, 12, 12)
        assert john.to_mongo(update=True) == {
            '$set': {'name': 'William Doe', 'birthday': datetime(1996, 12, 12)}}

    def test_dump(self):
        john = Student.build_from_mongo({
            'name': 'John Doe', 'birthday': datetime(1995, 12, 12), 'gpa': 3.0})
        assert john.dump() == {
            'name': 'John Doe',
            'birthday': '1995-12-12T00:00:00+00:00',
            'gpa': 3.0
        }


class TestConfig:

    def test_missing_schema(self):
        # No exceptions should occur

        class Doc1(Document):
            pass

        d = Doc1()
        assert isinstance(d.schema, Schema)

    def test_base_config(self):

        class Doc2(Document):
            pass

        assert Doc2.config['collection'] is None
        assert Doc2.config['lazy_collection'] is None
        assert Doc2.config['register_document'] is True

    def test_lazy_collection(self):

        def lazy_factory():
            return "fake_collection"

        class Doc3(Document):
            pass

            class Config:
                lazy_collection = lazy_factory

        assert Doc3.config['collection'] is None
        assert Doc3.config['lazy_collection'] is lazy_factory
        # Try to do the dereferencing
        assert Doc3.collection == "fake_collection"
        d = Doc3()
        assert d.collection == "fake_collection"

    def test_inheritance(self):

        class Doc4(Document):

            class Config:
                collection = "fake_collection"
                register_document = False


        class DocChild4(Doc4):

            class Config:
                collection = "fake_collection_2"

        assert Doc4.config['collection'] == "fake_collection"
        assert Doc4.config['lazy_collection'] is None
        assert Doc4.config['register_document'] is False
        assert DocChild4.config['collection'] == "fake_collection_2"
        assert DocChild4.config['lazy_collection'] is None
        assert DocChild4.config['register_document'] is False
        assert DocChild4.collection == "fake_collection_2"



from .abstract import BaseWrappedData, ChangeTracker
from .data_proxy import DataProxy
from .exceptions import ValidationError
from .meta import MetaEmbeddedDocument


__all__ = ('EmbeddedDocument', 'List', 'Reference')


class EmbeddedDocument(ChangeTracker, DataProxy, metaclass=MetaEmbeddedDocument):

    class Config:
        register_document = False

    def __init__(self, **kwargs):
        schema = self.Schema()
        super().__init__(schema, data=kwargs)
        self.__dict__['_callback'] = None

    def has_changed(self):
        return bool(self._modified_data)

    def _trigger_change(self):
        if self._callback:
            self._callback()

    def change_tracker_connect(self, callback):
        self._callback = callback

    def __delitem__(self, name):
        self._trigger_change()
        super().__delitem__(name)

    def __setitem__(self, name, value):
        self._trigger_change()
        super().__setitem__(name, value)


class List(ChangeTracker, list):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._has_changed = False
        self._callback = None

    def has_changed(self):
        return self._has_changed

    def _trigger_change(self):
        self._has_changed = True
        if self._callback:
            self._callback()

    def change_tracker_connect(self, callback):
        self._callback = callback

    def append(self, *args, **kwargs):
        ret = super().append(*args, **kwargs)
        self._trigger_change()
        return ret

    def pop(self, *args, **kwargs):
        ret = super().pop(*args, **kwargs)
        self._trigger_change()
        return ret

    def clear(self, *args, **kwargs):
        ret = super().clear(*args, **kwargs)
        self._trigger_change()
        return ret

    def remove(self, *args, **kwargs):
        ret = super().remove(*args, **kwargs)
        self._trigger_change()
        return ret

    def reverse(self, *args, **kwargs):
        ret = super().reverse(*args, **kwargs)
        self._trigger_change()
        return ret

    def sort(self, *args, **kwargs):
        ret = super().sort(*args, **kwargs)
        self._trigger_change()
        return ret

    def extend(self, *args, **kwargs):
        ret = super().extend(*args, **kwargs)
        self._trigger_change()
        return ret


class Reference(BaseWrappedData):

    def __init__(self, document_cls, pk):
        self.document_cls = document_cls
        self.pk = pk
        self._document = None

    def io_fetch(self):
        # Sync version
        if not self._document:
            if self.pk is None:
                raise ReferenceError('Cannot retrieve a None Reference')
            self._document = self.document_cls.find_one(self.pk)
            if not self._document:
                raise ValidationError(
                    'Reference not found for document %s.' % self._document_cls.__name__)
        return self._document



from abc import ABCMeta, abstractstaticmethod, abstractmethod


class AbstractDal(metaclass=ABCMeta):

    @abstractstaticmethod
    def is_compatible_with(collection):
        pass

    @abstractmethod
    def reload(self, doc):
        pass

    @abstractmethod
    def commit(self, doc, io_validate_all=False):
        pass

    @abstractmethod
    def delete(self, doc):
        pass

    @abstractmethod
    def find_one(self, doc_cls, *args, **kwargs):
        pass

    @abstractmethod
    def find(self, doc_cls, *args, **kwargs):
        pass


class AbstractCursor:
    # TODO...
    pass


class ChangeTracker:

    def has_changed(self):
        raise NotImplemented()

    def change_tracker_connect(self, callback):
        raise NotImplemented()


class BaseWrappedData:

    def to_mongo(self):
        raise NotImplemented()

    def from_mongo(self, **kwargs):
        raise NotImplemented()

    def dump(self):
        raise NotImplemented()

    def load(self, data):
        raise NotImplemented()
